﻿#include <algorithm>
#include <stdlib.h>
#include <unordered_set>
#include <iterator>
#include <map>
#include "MeshLoader.h"
#include <fstream>

std::set<Node> MeshLoader::getNodes() 
{
	return nodes;
}

std::set<FiniteElement> MeshLoader::getFiniteElements() 
{
	return elements;
}

std::multiset<BoundaryFiniteElement> MeshLoader::getBoundaryFiniteElements() 
{
	return boundaryFEs;
}


//!!! Контенеры-результаты должны содержать только ID узлов или КЭ


std::vector<FiniteElement> MeshLoader::getElementsByID(const int id1, const int id2, const int id3) //возможность нескольких узлов - контейнер
{
	/*std::vector<FiniteElement> result;
	auto current_iter = elements.begin();
	auto last_iter = elements.end();
	FiniteElement current_element;
	while (current_iter != last_iter) 
	{
		current_iter = std::find_if(current_iter, last_iter, [id1, id2, id3, &current_element]() {
			return std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id1) != current_element.nodesID.end()
				&& std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id2) != current_element.nodesID.end()
				&& std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id3) != current_element.nodesID.end(); });
		if (current_iter != last_iter) 
		{
			result.push_back(*current_iter);
			++current_iter;
		}
	}
	return result;*/
	std::vector<FiniteElement> result;
	auto current_iter = elements.begin();
	while (current_iter != elements.end())
	{
		current_iter = std::find_if(current_iter, elements.end(), [id1, id2, id3](const FiniteElement& current_element) {
			return std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id1) != current_element.nodesID.end()
				&& std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id2) != current_element.nodesID.end()
				&& std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id3) != current_element.nodesID.end(); });
		
		if (current_iter != elements.end())
		{
			result.push_back(*current_iter);
			current_iter++;
		}
	}
	return result;//O(n)
}

std::vector<FiniteElement> MeshLoader::getElementsByEdge(const int id1, const int id2) //возможность нескольких КЭ 
{
	/*std::vector<FiniteElement> result;
	auto current_iter = elements.begin();
	auto last_iter = elements.end();
	FiniteElement current_element;
	while (current_iter != last_iter) 
	{
		current_iter = std::find_if(current_iter, last_iter, [id1, id2, &current_element]() {
			return std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id1) != current_element.nodesID.end()
				&& std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id2) != current_element.nodesID.end(); });
		if (current_iter != last_iter) 
		{
			result.push_back(*current_iter);
			++current_iter;
		}
	}
	return result;*/
	std::vector<FiniteElement> result;
	std::copy_if(elements.begin(), elements.end(), std::back_inserter(result), [id1, id2](const FiniteElement& current_element) {
		return std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id1) != current_element.nodesID.end()
			&& std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id2) != current_element.nodesID.end(); });
	return result;
	//O(n)
}

std::vector<FiniteElement> MeshLoader::getElementsByOneID(const int id1) 
{
	/*std::vector<FiniteElement> result;
	auto current_iter = elements.begin();
	auto last_iter = elements.end();
	FiniteElement current_element;
	while (current_iter != last_iter) 
	{
		current_iter = std::find_if(current_iter, last_iter, [id1,&current_element]() {
			return std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id1) != current_element.nodesID.end(); });
		if (current_iter != last_iter) 
		{
			result.push_back(*current_iter);
			++current_iter;
		}
	}
	return result;*/
	std::vector<FiniteElement> result;
	std::copy_if(elements.begin(), elements.end(), std::back_inserter(result), [id1](const FiniteElement& current_element) {
		return std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id1) != current_element.nodesID.end(); });
	return result;
	//O(n)
}

std::vector<Node> MeshLoader::getNodesByBoundary(int boundaryID) 
{
	/*std::set<Node> result;//оператор == определён
	std::for_each(boundaryFEs.begin(), boundaryFEs.end(), [&](const BoundaryFiniteElement& current_element) {
		if (current_element.boundaryID == boundaryID)
		{
			std::for_each(current_element.nodesID.begin(), current_element.nodesID.end(), [&]() {
				auto node_by_ID = std::find_if(nodes.begin(), nodes.end(), [&current_element](Node n) { return n.nodeID == current_element; });
				result.insert(*node_by_ID);
			});
			std::copy_if(nodes.begin(), nodes.end(), std::inserter(result, result.begin()), [](Node n) {
				
				});
			/*for (const auto& it : current_element.nodesID)
			{
				auto node_by_ID = std::find_if(nodes.begin(), nodes.end(), [it](Node n) { return n.nodeID == it; });
				result.insert(*node_by_ID);
			}
		};
	});*/
	std::vector<BoundaryFiniteElement> result;
	
	std::vector<Node> final_result; //!!! Не верный тип контейнера
	
	BoundaryFiniteElement tmp;
	tmp.boundaryID == boundaryID;
	typedef std::multiset<BoundaryFiniteElement>::iterator It;
	std::pair<It, It> range = boundaryFEs.equal_range(tmp);
	if (*(range.first) == tmp)
	{
		std::copy(range.first, range.second, std::back_inserter(result));
	}
	for_each(result.begin(), result.end(), [&](const BoundaryFiniteElement& elem) {
		for (int i = 0; i < 3; ++i)
		{
			Node nmp;
			nmp.nodeID = elem.nodesID[i];
			auto nodeIter = nodes.find(nmp);
			final_result.push_back(*nodeIter); //!!! Где гарантия, что узлы не будут дублироваться?
		}
<<<<<<< HEAD
		});//O(n*log(n))
=======
		});//O(log(n))
	
	//!!! В целом что-то сложно - можно проще.
		
>>>>>>> 8d510920027e6e5db93a76f67a421ab4702ac30d
	return final_result;
}

std::vector<FiniteElement> MeshLoader::getElementsByBoundary(int surfaceID) 
{
	/*std::set<FiniteElement> result;
	auto current_iter = boundaryFEs.begin();
	auto last_iter = boundaryFEs.end();
	while (current_iter != last_iter) 
	{
		current_iter = std::find_if(current_iter, last_iter, [boundaryID](BoundaryFiniteElement current_element) {
			return current_element.boundaryID == boundaryID; });
		if (current_iter != last_iter) 
		{
			auto element_by_ID = std::find_if(elements.begin(), elements.end(), [current_iter](FiniteElement n) {
				return n.elementID == current_iter->elementID; });
			result.insert(*element_by_ID);
			++current_iter;
		}
	}*/
	std::vector<FiniteElement> result;//vector?
	std::copy_if(boundaryFEs.begin(), boundaryFEs.end(), std::back_inserter(result), [surfaceID](const BoundaryFiniteElement& current_element) {
		return surfaceID == current_element.surfaceID;
		});
	return result;//O(n)
}

std::vector<BoundaryFiniteElement> MeshLoader::getsboundaryFEsByBoundary(int boundaryID)//что поменять?
{
	/*std::vector<BoundaryFiniteElement> result;
	auto current_iter = boundaryFEs.begin();
	auto last_iter = boundaryFEs.end();
	std::copy_if(boundaryFEs.begin(), boundaryFEs.end(), std::back_inserter(result), [boundaryID](const BoundaryFiniteElement& current_element) {
			return current_element.boundaryID == boundaryID;
		});
	{
		current_iter = std::find_if(current_iter, last_iter, [boundaryID](BoundaryFiniteElement current_element) {
			return current_element.boundaryID == boundaryID; });
		if (current_iter != last_iter)
		{
			result.push_back(*current_iter);
			++current_iter;
		}
	}
	return result;*/
	/*std::vector<BoundaryFiniteElement> result;
	auto current_iter = boundaryFEs.begin();
	auto last_iter = boundaryFEs.end();
	while (current_iter != last_iter) 
	{
		current_iter = std::find_if(current_iter, last_iter, [boundaryID](BoundaryFiniteElement current_element) {
			return current_element.boundaryID == boundaryID; });
		if (current_iter != last_iter) 
		{
			result.push_back(*current_iter);
			++current_iter;
		}
	}
	return result;*/
	std::vector<BoundaryFiniteElement> result;//vector?
	BoundaryFiniteElement tmp;
	tmp.boundaryID == boundaryID;
	typedef std::multiset<BoundaryFiniteElement>::iterator It;
	std::pair<It, It> range = boundaryFEs.equal_range(tmp);
	if (*(range.first) == tmp)
	{
		std::copy(range.first, range.second, std::back_inserter(result));
	}
	return result;//O(log(n))
}

void MeshLoader::insertCenterNode() {
	auto edges = deleteEdgesDublicates();
	for (auto it = elements.begin(); it != elements.end(); it++) 
    {
		std::vector<int> current_nodes(it->nodesID);
		for (auto i = current_nodes.begin(); i < current_nodes.end(); ++i)
			for (auto j = i + 1; j < current_nodes.end(); ++j) 
			{
				Edge current_edge(*i, *j, 0);
				auto founded_edge = edges.find(current_edge);
				if (founded_edge == edges.cend()) {
					Node left_node = *getNodeByID(*i);
					Node right_node = *getNodeByID(*j);
					int center_id_ = getLastNodeID() + 1;
					Node center_node(center_id_, (left_node.x + right_node.x) / 2,
						(left_node.y + right_node.y) / 2, (left_node.z + right_node.z) / 2, false);

					current_edge.updateCenter(center_id_);
					nodes.insert(center_node);
					edges.insert(current_edge);
					it->nodesID.push_back(center_id_);//добавил существующий узел в элемент
				}
				//!!! А если ребро есть, то существующий ID узла не надо в КЭ добавить?
			}
	}

	for (auto it = boundaryFEs.begin(); it != boundaryFEs.end(); it++)
	{
		std::vector<int> current_nodes(it->nodesID);
		for (auto i = current_nodes.begin(); i < current_nodes.end() - 1; ++i)//
			for (auto j = i + 1; j < current_nodes.end(); ++j) 
			{
				Edge current_edge(*i, *j, 0);
				auto founded_edge = edges.find(current_edge);
				if (founded_edge != edges.cend())
				{
					it->nodesID.push_back(founded_edge->center_id);
				}
			}
	}
}//O(n^2)



std::set<Node>::iterator MeshLoader::getNodeByID(int id) 
{
	Node tmp;
	tmp.nodeID = id;
	return nodes.find(tmp);
}

int MeshLoader::getLastNodeID() 
{
	return nodes.rbegin()->nodeID;
}



std::map<int,std::vector<Node>> MeshLoader::getNodeNeighbours() 
{
	std::map<int, std::vector<Node>> result_out; //!!! Не подходящий контейнер std::vector
	
	//!!! Слишком сложная реализация - упростить
	std::for_each(nodes.begin(), nodes.end(), [&](const Node& current_node) {
		std::set<int> result_in;
		std::vector<FiniteElement> element_vec = getElementsByOneID(current_node.nodeID);
		std::vector<Node> NodeNeighbors;
		
		std::for_each(element_vec.begin(), element_vec.end(), [&result_in](const FiniteElement& current_elem) {
			result_in.insert(current_elem.nodesID.begin(), current_elem.nodesID.end());
			});

		result_in.erase(current_node.nodeID);

		std::for_each(result_in.begin(), result_in.end(), [&](int curr_nodeID) {
			NodeNeighbors.push_back(*getNodeByID(curr_nodeID));
			});
		result_out.emplace(current_node.nodeID, NodeNeighbors);
	});
	return result_out;
}//O(n^2)


void MeshLoader::Print_Data() 
{
	
	std::cout << getNodes() << std::endl
		<< getFiniteElements() << std::endl
		<< getBoundaryFiniteElements() << std::endl;
}

std::unordered_set <Edge, std::hash<Edge>> MeshLoader::deleteEdgesDublicates()
{
    //!!! Не понятный метод. unordered_set и так дубликаты не хранит. См. замечание к hash-функции.
    
	std::unordered_set <Edge, std::hash<Edge>> result;
	auto current_iter = edges.begin();
	auto last_iter = edges.end();
	for (auto& i : edges)
	{
		for (auto it=++current_iter; it != edges.end(); (it->first_node_id == i.last_node_id && it->last_node_id == i.first_node_id) ? it = edges.erase(it) : ++it)
		{
			result.insert(*it);
		}
	}
	return result;
}//O(n)

std::vector<BoundaryFiniteElement> MeshLoader::getTrianglesAndNodes(int surfaceID)
{
	std::vector<BoundaryFiniteElement> result;//vector?
	std::copy_if(boundaryFEs.begin(), boundaryFEs.end(), std::back_inserter(result), [surfaceID](const BoundaryFiniteElement& current_element) {
		return surfaceID == current_element.surfaceID;
		});
	auto square = [](Node A, Node B, Node C)
	{
		double a = sqrt((C.x - B.x) * (C.x - B.x) + (C.y - B.y) * (C.y - B.y) + (C.z - B.z) * (C.z - B.z));
		double b = sqrt((C.x - A.x) * (C.x - A.x) + (C.y - A.y) * (C.y - A.y) + (C.z - A.z) * (C.z - A.z));
		double c = sqrt((B.x - A.x) * (B.x - A.x) + (B.y - A.y) * (B.y - A.y) + (B.z - A.z) * (B.z - A.z));
		double p = (a + b + c)/2;
		return sqrt(p * (p - a) * (p - b) * (p - c));
	};
	std::ranges::sort(result, [&](const BoundaryFiniteElement& a, const BoundaryFiniteElement& b)
		{
			Node X1 = *getNodeByID(a.nodesID[0]);
			Node Y1 = *getNodeByID(a.nodesID[1]);
			Node Z1 = *getNodeByID(a.nodesID[2]);

			Node X2 = *getNodeByID(b.nodesID[0]);
			Node Y2 = *getNodeByID(b.nodesID[1]);
			Node Z2 = *getNodeByID(b.nodesID[2]);

			return square(X1,Y1,Z1) > square(X2,Y2,Z2); 
		});
	std::ofstream triangle("triangle.txt", std::ofstream::out | std::ofstream::trunc);

	std::ostream_iterator<BoundaryFiniteElement> os(triangle, " ");
	int counter = 0;
	ranges::copy_if(result, std::ostream_iterator<BoundaryFiniteElement>(triangle, "\n"),
		[&counter](const auto& x) 
		{ 
		counter++;
		return ((counter - 1) % 20) >= 1; 
		});
	triangle.close();

	return result;//O(n)
}